import re

PATTERNS = {
    'phone': r"""^[+]*[(]{0,1}[0-9]{0,4}[)]{0,1}[-\s\./0-9]*$""",
    'email': r"""(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])""",
}
REGEXPS = {name: re.compile(pattern) for name, pattern in PATTERNS.items()}


def required(instance, field, value, message=None):
    if not value:
        raise ValueError(message or f'{field} is required')


def phone_number(instance, field, value, message=None):
    if re.match(REGEXPS['phone'], value) is None:
        raise ValueError(message or f'{field} does not match the valid pattern.')


def email(instance, field, value, message=None):
    if re.match(REGEXPS['email'], value) is None:
        raise ValueError(message or f'{field} does not match the valid pattern.')



import json, sys
import dataclasses
from .util import JSONEncoder, classproperty
from . import validators


class Model:
    """
    Create a model that builds on dataclasses to provide:
    * automatic post-init
    * converters 
    * validators
    * input from data
    * output to data
    """

    CONVERTERS = {}
    VALIDATORS = {}

    @classproperty
    def FIELDS(cls):
        return cls.__dataclass_fields__

    @classmethod
    def load(cls, data, keys=None):
        """pull field values out of a data source, ignoring data that is not in the model"""
        if not keys:
            keys = cls.__dataclass_fields__.keys()
        return cls(**{key: data[key] for key in keys if key in data})

    def __post_init__(self):
        """cast the input field values to the declared type, and then run any declared converters"""
        for field in self.__dataclass_fields__:
            try:
                value = getattr(self, field)
                if value:
                    # if converters are specified, use those;
                    # otherwise, cast the value to the field_type
                    field_metadata = self.__dataclass_fields__[field].metadata

                    converters = (field_metadata.get('converters') or []) + (
                        self.CONVERTERS.get(field) or []
                    )
                    if converters:
                        for converter in converters:
                            setattr(self, field, converter(value))
                    # else:
                    #     field_type = self.__dataclass_fields__[field].type
                    #     # typing.List, etc. have __origin__ and __args__
                    #     if field_type.__dict__.get('__origin__'):
                    #         origin_type = field_type.__dict__['__origin__']
                    #         if field_type.__dict__.get('__args__'):
                    #             arg_types = field_type.__dict__['__args__']
                    #             if len(arg_types) == 1:
                    #                 setattr(self, field, origin_type(arg_types[0](value)))
                    #             else:
                    #                 setattr(
                    #                     self,
                    #                     field,
                    #                     origin_type(
                    #                         arg_types[i](value[i]) for i in range(len(arg_types))
                    #                     ),
                    #                 )
                    #         else:
                    #             setattr(self, field, origin_type(value))
                    #     else:
                    #         setattr(self, field, field_type(value))
            except ValueError as exc:
                if not exc.args:
                    exc.args = ('',)
                exc.args = (f'{self.__class__.__name__}.{field}: ' + ' '.join(exc.args),)
                raise

    def dict(self, nulls=True):
        return {k: v for k, v in dataclasses.asdict(self).items() if nulls or v}

    def json(self, indent=None, nulls=True):
        return json.dumps(self.dict(nulls=nulls), indent=indent, cls=JSONEncoder)

    def json_dict(self, indent=None, nulls=True):
        """return a plain-jsonable dict"""
        return json.loads(self.json(indent=indent, nulls=nulls))

    def keys(self, nulls=True):
        return self.dict(nulls=nulls).keys()

    def values(self, nulls=True):
        return iter([v for k, v in self.dict(nulls=nulls).items()])

    def items(self, nulls=True):
        return self.dict(nulls=nulls).items()

    def __iter__(self):
        for key in self.keys():
            yield key

    def is_valid(self):
        return not (bool(self.errors))

    def assert_valid(self):
        errors = self.errors
        if bool(errors) is True:
            raise ValueError("Validation failed: %s" % json.dumps(errors, indent=2))

    @property
    def pk(self):
        return {
            field: getattr(self, field)
            for field in self.__dataclass_fields__
            if self.__dataclass_fields__[field].metadata.get('pk')
            or (hasattr(self, 'PK') and field in self.PK)
        }

    @property
    def errors(self):
        """returns a dict of validation errors, keyed to attribute names"""
        validation_errors = {}
        for field in self.__class__.__dataclass_fields__:
            field_metadata = self.__dataclass_fields__[field].metadata
            validators = (field_metadata.get('validators') or []) + (
                self.VALIDATORS.get(field) or []
            )
            if validators:
                value = getattr(self, field)
                err = []
                for validator in validators:
                    try:
                        validator(self, field, value)
                    except ValueError:
                        err.append(str(sys.exc_info()[1]))
                if len(err) > 0:
                    validation_errors[field] = err
        return validation_errors



import dateparser
from datetime import datetime, date
from dataclasses import dataclass
from datamodels import Model


@dataclass
class Person(Model):
    name: str
    title: str = None
    birthdate: date = None

    class CONVERTERS:
        def birthdate(value):
            return (
                dateparser.parse(value).date()
                if not isinstance(value, (datetime, date))
                else value
            )

    class VALIDATORS:
        def title(instance, field, value):
            print(instance, field, value)
            values = {'Mr', 'Ms', 'Mrs', 'Dr'}
            if value is not None and value not in values:
                raise ValueError(f'{field} must be one of {values}.')



import dataclasses
from . import validators


# NOT READY FOR USE

def Field(
    *, default=None, pk=False, required=False, compare=True, converter=None, validator=None, **kw
):
    _metadata = {**kw}
    if isinstance(default, type) or str(type(default) == "<class 'function'>"):
        _default_factory = default
        _default = dataclasses.MISSING
    else:
        _default_factory = dataclasses.MISSING
        _default = default

    if pk:
        _repr = True
        _hash = True
        _metadata['pk'] = True
    else:
        _repr = False
        _hash = False
        _metadata['pk'] = False

    _compare = compare
    _init = True

    if not converter:
        _metadata['converters'] = []
    elif not isinstance(converter, list):
        _metadata['converters'] = [converter]
    else:
        _metadata['converters'] = converter

    if not validator:
        _metadata['validators'] = []
    elif not isinstance(validator, list):
        _metadata['validators'] = [validator]
    else:
        _metadata['validators'] = validator

    if required:
        _metadata['validators'].insert(0, validators.required)

    return dataclasses.field(
        default=_default,
        default_factory=_default_factory,
        init=_init,
        repr=_repr,
        hash=_hash,
        compare=_compare,
        metadata=_metadata,
    )



from datetime import date, datetime, time
import json
import re


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime) or isinstance(obj, date) or isinstance(obj, time):
            return str(obj)
        return super().default(obj)


class classproperty(property):
    def __get__(self, obj, objtype=None):
        return super(classproperty, self).__get__(objtype)

    def __set__(self, obj, value):
        super(classproperty, self).__set__(type(obj), value)

    def __delete__(self, obj):
        super(classproperty, self).__delete__(type(obj))


def nameify(string, ascii=False, sep='_'):
    """return an XML name (hyphen-separated by default, initial underscore if non-letter)"""
    s = camelsplit(string)  # immutable
    s = hyphenify(s, ascii=ascii).replace('-', sep)
    if len(s) == 0 or re.match("[A-Za-z_]", s[0]) is None:
        s = "_" + s
    return s


def camelsplit(string):
    """Turn a CamelCase string into a string with spaces"""
    s = str(string)
    for i in range(len(s) - 1, -1, -1):
        if i != 0 and (
            (s[i].isupper() and s[i - 1].isalnum() and not s[i - 1].isupper())
            or (s[i].isnumeric() and s[i - 1].isalpha())
        ):
            s = s[:i] + ' ' + s[i:]
    return s.strip()


def hyphenify(string, ascii=False):
    s = str(string)
    s = re.sub("""['"\u2018\u2019\u201c\u201d]""", '', s)  # quotes
    s = re.sub(r'(?:\s|%20)+', '-', s)  # whitespace
    if ascii == True:  # ASCII-only
        s = s.encode('ascii', 'xmlcharrefreplace').decode('ascii')  # use entities
    s = re.sub("&?([^;]*?);", r'.\1-', s)  # entities
    s = s.replace('#', 'u')
    s = re.sub(r"\W+", '-', s).strip(' -')
    return s
